// ServerFrame.java
//
// Informatics 45 Spring 2010
// Code Example: GUI with Simple Sockets
//
// This is the server GUI.  You'll notice that much of what's going on here
// is familiar: creating UI components, laying them out using GridBagLayout,
// attaching an ActionListener to a button.  I haven't explained those parts,
// except where we're doing something new that we haven't done before.
//
// The general idea of how the GUI interacts with the socket communication
// handling is this:
//
// * When the GUI is ready to listen for a connection, it creates a
//   ServerProtocol object.  This object knows all the details of how to
//   listen for connections, how to create sockets, and what the communication
//   protocol is.
//
// * ServerProtocols do the majority of their work in a separate thread.  They
//   communicate back to the GUI in a familiar way: the GUI registers itself
//   as a "server protocol listener," then receives method calls from the
//   ServerProtocol.
// 
//   The key difference, though, is that the method calls from the
//   ServerProtocol are coming in on the server thread -- the one handling the
//   sockets -- and not the UI thread.  That means that they have to do
//   something special: place an event on the Swing event queue, so that the
//   UI-related changes are done on the UI thread (a requirement of Swing).
//
// More details below.

package inf45.spring2010.examples.sockets3.gui;

import inf45.spring2010.examples.sockets3.protocol.ServerProtocol;
import inf45.spring2010.examples.sockets3.protocol.ServerProtocolListener;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;


// Since the frame can act as a listener, this class not only extends from
// JFrame but also implements ServerProtocolListener.

public class ServerFrame extends JFrame implements ServerProtocolListener
{
	// These are the parts of the user interface that are accessed or
	// can change after the UI has been built, so we store them in
	// fields.
	private JTextField nameField;
	private JTextField portField;
	private JButton listenButton;
	private DefaultListModel resultsListModel;

	// We'll keep a reference to the current ServerProtocol, pointing it to
	// null if there isn't one.  This will be important, since it will
	// allow us to shut the current ServerProtocol down (if there is one)
	// when the window is closing.
	private ServerProtocol serverProtocol;
	
	
	public ServerFrame()
	{
		setTitle("Informatics 45 GUIs and Sockets Example (Server)");
		setSize(400, 600);
		setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
		
		buildUI();

		// Initially, there is no acceptor; we only start to wait for a
		// connection after the "Listen" button is clicked.
		serverProtocol = null;
	}
	
	
	private void buildUI()
	{
		GridBagLayout layout = new GridBagLayout();
		getContentPane().setLayout(layout);
		
		JLabel nameLabel = new JLabel("Your name:");
		getContentPane().add(nameLabel);
		layout.setConstraints(
			nameLabel,
			new GridBagConstraints(
				0, 0, 1, 1, 0.0, 0.0,
				GridBagConstraints.WEST, GridBagConstraints.NONE,
				new Insets(5, 5, 5, 5), 0, 0));
		
		nameField = new JTextField();
		getContentPane().add(nameField);
		layout.setConstraints(
			nameField,
			new GridBagConstraints(
				1, 0, 1, 1, 1.0, 0.0,
				GridBagConstraints.CENTER, GridBagConstraints.HORIZONTAL,
				new Insets(5, 5, 5, 5), 0, 0));
		
		JLabel portLabel = new JLabel("Listen port:");
		getContentPane().add(portLabel);
		layout.setConstraints(
			portLabel,
			new GridBagConstraints(
				0, 1, 1, 1, 0.0, 0.0,
				GridBagConstraints.WEST, GridBagConstraints.NONE,
				new Insets(5, 5, 5, 5), 0, 0));
		
		portField = new JTextField();
		getContentPane().add(portField);
		layout.setConstraints(
			portField,
			new GridBagConstraints(
				1, 1, 1, 1, 1.0, 0.0,
				GridBagConstraints.CENTER, GridBagConstraints.HORIZONTAL,
				new Insets(5, 5, 5, 5), 0, 0));
		
		listenButton = new JButton("Listen");
		listenButton.addActionListener(
			new ActionListener()
			{
				public void actionPerformed(ActionEvent e)
				{
					listen();
				}
			});
		getContentPane().add(listenButton);
		layout.setConstraints(
			listenButton,
			new GridBagConstraints(
				1, 2, 1, 1, 0.0, 0.0,
				GridBagConstraints.EAST, GridBagConstraints.NONE,
				new Insets(5, 5, 5, 5), 0, 0));
		
		JLabel resultsLabel = new JLabel("Results:");
		getContentPane().add(resultsLabel);
		layout.setConstraints(
			resultsLabel,
			new GridBagConstraints(
				0, 3, 1, 1, 0.0, 0.0,
				GridBagConstraints.WEST, GridBagConstraints.NONE,
				new Insets(5, 5, 5, 5), 0, 0));

		resultsListModel = new DefaultListModel();
		JList resultsList = new JList(resultsListModel);
		JScrollPane resultsScrollPane = new JScrollPane(resultsList);
		getContentPane().add(resultsScrollPane);
		layout.setConstraints(
			resultsScrollPane,
			new GridBagConstraints(
				0, 4, 2, 1, 1.0, 1.0,
				GridBagConstraints.CENTER, GridBagConstraints.BOTH,
				new Insets(5, 5, 5, 5), 0, 0));
	}
	
	
	// This method is called when the "Listen" button is clicked.  It
	// starts by doing what you might call "form validation."  The user
	// has filled out a form consisting of their name and the port they'd
	// like to listen on; the UI, in turn, checks to make sure that's
	// been done appropriately, by checking for three problems and gently
	// prompting the user to fix them if they have occurred.
	//
	// (1) Did the name get specified or was it left empty?  We'll trim
	//     the extra space off of the beginning and end of what the user
	//     specified; if there's nothing left, the user didn't enter a
	//     name.
	//
	// (2) Did the user specify an integer for the listen port?  If not
	//     (e.g., they typed "Boo" instead), the user should be prompted to
	//     enter one.
	//
	// (3) Did the user specify a valid port number?  Valid port numbers
	//     on the Internet range from 0-65535.  If the user typed a number
	//     for the port, but it wasn't valid, we should prompt the user
	//     to fix it.
	//
	// It's important that we validate user input, so that we can display
	// gentle, targeted error messages when the input is incorrect, rather
	// that letting the program spectacularly blow up trying to process it
	// (which can lead to all kinds of problems, including security issues
	// when things go really wrong).
	//
	// Once we've validated the user input, we'll make a few changes to the
	// user interface, then create and start up a ServerProtocol.
	private void listen()
	{
		String name = getNameFromField();
		int port = getPortFromField();
		
		if (name == null || port == -1)
		{
			return;
		}
		
		listenButton.setEnabled(false);
		
		resultsListModel.addElement("Listening for connection");
		
		serverProtocol = new ServerProtocol(name, port);
		serverProtocol.addServerProtocolListener(this);
		
		Thread listenerThread = new Thread(serverProtocol);
		listenerThread.start();
	}
	
	
	private String getNameFromField()
	{
		String name = nameField.getText().trim();
		
		if (name.length() == 0)
		{
			JOptionPane.showMessageDialog(
				this, "Name must be specified.",
				"Validation Error",
				JOptionPane.ERROR_MESSAGE);
			
			return null;
		}
		
		return name;
	}
	
	
	private int getPortFromField()
	{
		int port;
		
		try
		{
			port = Integer.parseInt(portField.getText().trim());
			
			if (port < 0 || port > 65535)
			{
				JOptionPane.showMessageDialog(
					this, "Listen port must be between 0 and 65535.",
					"Validation Error",
					JOptionPane.ERROR_MESSAGE);
				
				return -1;
			}
		}
		catch (NumberFormatException e)
		{
			JOptionPane.showMessageDialog(
				this, "Listen port must be a number.",
				"Validation Error",
				JOptionPane.ERROR_MESSAGE);
			
			return -1;
		}
		
		return port;
	}
	
	
	// This is the listener method that is called when the server protocol
	// has succeeded in accepting a connection.  The protocol implementation
	// gives us back the client's address, so we know where the client is
	// calling from.  We'll take that information and add it to the list
	// of results shown in the GUI.
	//
	// Notice, though, that the method doesn't actually do the work itself.
	// Instead, it places an event on the Swing event queue -- by calling
	// EventQueue.invokeLater().  This is a way of telling the UI thread:
	// "Here's a job I want you to do when you get around to it."  The way
	// you specify what job it should do is to provide it a Runnable object;
	// Runnable objects have (at least) one method called run(), which does
	// the actual work.
	//
	// You may be wondering why the parameter here is marked "final."
	// This is because of a limitation of Java.  Where we created a new
	// Runnable object and implemented it, we've created something called
	// an "anonymous inner class": a new class with no name (hence the
	// term "anonymous") that implements the Runnable interface.  Anonymous
	// inner classes have access to the local variables and parameters of
	// the methods that create them, but only if the local variables or
	// parameters are guaranteed never to change.  You can guarantee that
	// something will never change, in Java, by marking it "final."  So
	// we mark the parameters "final" so that they can be used in the
	// anonymous inner class, so that the UI thread can ultimately make
	// use of them.  Phew!
	public void connectionAccepted(final String clientAddress)
	{
		EventQueue.invokeLater(
			new Runnable()
			{
				public void run()
				{
					resultsListModel.addElement("Connection accepted from " + clientAddress);
				}
			});
	}

	
	// This is the method that is called by the server protocol when the HELLO
	// message is received from the client.
	public void helloReceived(final String otherName)
	{
		EventQueue.invokeLater(
			new Runnable()
			{
				public void run()
				{
					resultsListModel.addElement("Received HELLO from " + otherName);
				}
			});
	}

	
	// This is the method that is called by the server protocol when the
	// connection is closed.  At this point, we'll re-enable the Listen
	// button, so that it can be pressed again.
	public void connectionClosed()
	{
		EventQueue.invokeLater(
			new Runnable()
			{
				public void run()
				{
					resultsListModel.addElement("Connection closed");
					listenButton.setEnabled(true);
				}
			});
	}

	
	// This is the method that is called by the server protocol when
	// something goes wrong during the conversation with the client
	// (e.g., the client is not following the appropriate protocol).
	// At this point, we'll re-enable the Listen button, so that it
	// can be pressed again.
	public void protocolFailed(final Exception failureReason)
	{
		EventQueue.invokeLater(
			new Runnable()
			{
				public void run()
				{
					resultsListModel.addElement(
						"Protocol failure: " + failureReason.getMessage());
						
					listenButton.setEnabled(true);
				}
			});
	}
	
	
	// The dispose() method is called as the window is being destroyed after
	// it is closed.  Java programs continue to run as long as any of their
	// threads are still running.  Since the ServerProtocol, if there is one,
	// is doing its work in a separate thread, that means the program will
	// continue to run even if the GUI window has been dismissed -- unless we
	// do something to make sure that the ServerProtocol is killed off if it's
	// still running.  This method does that by calling the ServerProtocol's
	// close() method, which kills it off if it's in the process of running.
	public void dispose()
	{
		if (serverProtocol != null)
		{
			// We remove ourselves as a listener, since we're only interested in
			// having the ServerProtocol stop; it will complain to us that it
			// failed, but we don't care.  (And, furthermore, even if we did
			// care, there would be no window left to respond to the event once
			// it's been dismissed, so we *have* to not care.)
			serverProtocol.removeServerProtocolListener(this);
			serverProtocol.close();
		}

		// This says "Now do what you normally do when windows are being
		// disposed."  (If we don't do this, the window won't actually close.)
		super.dispose();
	}
}
