// ServerProtocol.java
//
// Informatics 45 Spring 2010
// Code Example: GUI with Simple Sockets
//
// This is the ServerProtocol, whose job is to wait for a connection, then
// have a full conversation, according to our protocol, with the client who
// connected.  This work can potentially take a long time, so we'll run it
// in a separate thread, then notify "listeners" when interesting things
// happen -- the way that UI components notify our listeners when they are
// interacted with.

package inf45.spring2010.examples.sockets3.protocol;

import java.io.IOException;
import java.io.PrintWriter;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.ArrayList;
import java.util.Scanner;


// A ServerProtocol object represents a task that can be executed in a separate
// thread, so it implements the Runnable interface.  Threads begin by calling
// the run() method on some object that implements Runnable; they continue
// executing until the run() method returns (or terminates with an exception).

public class ServerProtocol implements Runnable
{
	private String name;
	private int listenPort;
	private ArrayList<ServerProtocolListener> listeners;
	
	private ServerSocket serverSocket;
	private Socket socket;
	
	
	public ServerProtocol(String name, int listenPort)
	{
		this.name = name;
		this.listenPort = listenPort;
		listeners = new ArrayList<ServerProtocolListener>();
		
		// Once the ServerProtocol starts running, either or both of these
		// will be non-null; initially, though, there are no sockets.
		serverSocket = null;
		socket = null;
	}
	
	
	public void addServerProtocolListener(ServerProtocolListener listener)
	{
		listeners.add(listener);
	}
	
	
	public void removeServerProtocolListener(ServerProtocolListener listener)
	{
		if (listeners.contains(listener))
		{
			listeners.remove(listener);
		}
	}
	
	
	// This code is mostly pretty uninteresting, in the sense that it's not
	// much different than what we were doing in the console-mode examples.
	// It may take a very long time to run this code (since it waits
	// indefinitely, until it receives a connection), but it won't affect
	// our program negatively, since this code is run in a separate thread,
	// simultaneously with the code that handles the user interface.
	//
	// The two interesting, different things we're doing here are these.
	//
	// (1) Once we accept a connection from a client, we can find out what
	//     the client's address is by asking the socket -- by calling
	//     getInetAddress(), then converting the result to a string, so that
	//     we can eventually show it in the user interface.
	//
	// (2) We're usually a "finally" block at the end of our try/catch.  There
	//     are two possible outcomes in a try/catch: the try runs without any
	//     exceptions and that's it, or the try block throws an exception,
	//     which causes control to transfer to the appropriate catch handler.
	//     A finally block is code that will always run no matter what, either
	//     after the try completes successfully or after a catch handler
	//     executes (or even if no matching catch handler is found, in which
	//     case the finally block executes before we bail out of the method).
	//     A finally block is a great place to close our socket, because it's
	//     guaranteed to run no matter what happens: if things go well or
	//     something bad happens, at least we can be sure that the socket is
	//     closed.
	public void run()
	{
		serverSocket = null;
		socket = null;
		
		try
		{
			serverSocket = new ServerSocket(listenPort);
			socket = serverSocket.accept();

			fireConnectionAccepted(socket.getInetAddress().toString());
			
			Scanner socketIn = new Scanner(socket.getInputStream());

			PrintWriter socketOut = new PrintWriter(
				socket.getOutputStream(), true);
			
			String helloLine = socketIn.nextLine();
			
			if (helloLine.startsWith("HELLO "))
			{
				String otherName = helloLine.substring("HELLO ".length());
				fireHelloReceived(otherName);
				socketOut.println("OKAY " + name);
			}
			else
			{
				throw new ServerProtocolException("Client must send HELLO");
			}
		}
		catch (Exception e)
		{
			fireProtocolFailed(e);
		}
		finally
		{
			close();
			fireConnectionClosed();
		}
	}
	
	
	// This method can be called by a thread other than the server thread.
	// It closes the sockets, if they exist, which has the effect of killing
	// off the server thread -- because whatever it's doing with the sockets
	// will fail with an exception, which will be caught, and then the
	// server thread's run() method will return, ending the thread.  It's
	// important that we provide a public mechanism for closing out the
	// ServerProtocol so that we can be sure that the program will end --
	// because, remember, a Java program doesn't end until all of its threads
	// have died, even if main() has returned.
	public void close()
	{
		if (socket != null)
		{
			try
			{
				socket.close();
			}
			catch (IOException ee)
			{
			}
		}
		
		socket = null;
		
		if (serverSocket != null)
		{
			try
			{
				serverSocket.close();
			}
			catch (IOException ee)
			{
			}
		}
		
		serverSocket = null;
	}
	
	
	private void fireConnectionAccepted(String clientAddress)
	{
		for (ServerProtocolListener listener : listeners)
		{
			listener.connectionAccepted(clientAddress);
		}
	}
	
	
	private void fireHelloReceived(String otherName)
	{
		for (ServerProtocolListener listener : listeners)
		{
			listener.helloReceived(otherName);
		}
	}
	
	
	private void fireProtocolFailed(Exception failureReason)
	{
		for (ServerProtocolListener listener : listeners)
		{
			listener.protocolFailed(failureReason);
		}
	}
	
	
	private void fireConnectionClosed()
	{
		for (ServerProtocolListener listener : listeners)
		{
			listener.connectionClosed();
		}
	}
}
